package org.processmining.uncertaintymarker.algorithms;

import java.util.ArrayList;

import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.deckfour.xes.model.impl.XAttributeMapImpl;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetImpl;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.uncertaintymarker.models.TransitionData;

/**<p>The process discovery step is the third out of four steps performed by this plugin. 
 * It is performed after {@link TraceAlignment} and before {@link NetAttachment}.</p>
 * 
 * <p>This class implements a simple Process Discovery Algorithm which I developed on my own. It works only with an aligned log as input.
 * Beginning with n=1, the algorithm looks at the n-th element of each trace, and groups them together into objects as {@link TransitionData}. 
 * If the frequency of a certain event is above a certain threshold that is set by the user as a parameter, 
 * the event will be included into the resulting Petrinet. Each element in such column n will get an arc from the same predecessor place, 
 * and one to the same successor place. This way the resulting net is a sequence of XOR-Connections.</p>
 * 
 * @author Oliver
 */


public class AlignedStepDiscovery {

	Petrinet net = new PetrinetImpl("");
	XLog log = null;
	
	int logLength;
	int logDepth;
	
	//collects different events in a column ( rested with each call of setcolumndistribution)
	ArrayList<TransitionData> transitions = new ArrayList<>();
	
	/**Main method of this class.
	 * The algorithm produces a Petrinet containing a sequence of XOR-connections of the events in the columns.
	 * For each column in the alignedLog, either one, or multiple events are included in the net in an XOR-Connection,
	 * depending on whether their relative frequency in this column is above the inclusionThreshold-parameter set by the user.
	 * If there is only the gap-sign event in a column which is meeting the condition, this column is skipped.
	 */
	public Petrinet discoverPetrinet(XLog alignedLog, double inclusionThreshold, PluginContext context){
		
		log = alignedLog;
		logLength = alignedLog.get(0).size();
		logDepth = alignedLog.size();
		
		//needed for trace alignment, not neccessary and anymore in this process discovery algorithm
//		NetAttachement.removeOccurenceNumbers(log,net);
		
		//adding start place
		Place currentPred = net.addPlace(""); 
		Place currSucc = null;
		
		for(int column=0;column< logLength;column++){
			
			//fill transitions for this column
			setColumnDistribution(column);
			
			ArrayList<TransitionData> subset = new ArrayList<>();
			
			//fillSubSetList based on Threshold	
			for(int i=0;i<transitions.size();i++){
				double frequency = transitions.get(i).getFrequency();
				double currRelativeFrequency = frequency/logDepth;
				if(currRelativeFrequency>=inclusionThreshold){
		        	subset.add(transitions.get(i));
		        }
		    }
			
			//only gap sign -> skip
			if( ( subset.size()==1 && subset.get(0).getLabel().equals("-") ) || subset.size()==0){ 
					continue;
			}
			//one or more transitions
			else{ 
				currSucc = net.addPlace("");
				for(int i=0;i<subset.size();i++){
					
					Transition currTrans = net.addTransition(subset.get(i).getLabel() );
					if(subset.get(i).getLabel().equals("-"))
						currTrans.setInvisible(true);
//						currTrans.getAttributeMap().put(AttributeMap.FILLCOLOR, Color.BLACK);
					
					net.addArc(currentPred, currTrans); // arc from predecessor place
					net.addArc(currTrans,currSucc); //arc to successor place					
				}
				currentPred = currSucc;
			}
		}
		
		
		//create markings and connections
		Place source = null;
		Place sink = null;
		for(Place p : net.getPlaces()){
			if(net.getInEdges(p).isEmpty())
				source = p; 
			if(net.getOutEdges(p).isEmpty())
				sink = p; 
		}
		
		
		Marking sourceM = new Marking();
		sourceM.add(source);
		Marking sinkM = new Marking();
		sinkM.add(sink);
		
		InitialMarkingConnection imc = new InitialMarkingConnection(net, sourceM);
		FinalMarkingConnection fmc = new FinalMarkingConnection(net, sinkM);
		
		
        context.getProvidedObjectManager().createProvidedObject(
                "Initial marking for " + net.getLabel(),
                sourceM, Marking.class, context);
        context.addConnection(imc);

        context.getProvidedObjectManager().createProvidedObject(
                "Final marking for " + net.getLabel(),
                sinkM, Marking.class, context);
        context.addConnection(fmc);
		
		
		return net;
	}
	
	/**Aggregate/Summarize all events contained in the specified column into objects of type {@link TransitionData} and 
	 * store them in the transitions list.
	 */
	public void setColumnDistribution(int column){
		
		transitions=new ArrayList<>();
		
		//fill list
		for(XTrace t: log){
			
			XAttributeMapImpl map = new XAttributeMapImpl(t.get(column).getAttributes());

			XAttributeLiteralImpl xAtt = (XAttributeLiteralImpl) map.get("concept:name");
			String currEvent = xAtt.getValue(); 
			
			boolean flag=true;
			for(int i=0;i<transitions.size();i++){ 
				TransitionData tr = transitions.get(i);
				if(tr.getLabel().equals(currEvent)){ //add row information to existing element
					tr.addItem(map,column);
					flag =false;
				}
			}
			if(flag){ //add new element
				TransitionData newTr2 = new TransitionData();
				newTr2.addItem(map,column);
				transitions.add(newTr2);
			}
		}
	}
}
