package org.processmining.uncertaintymarker.algorithms;

import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeContinuousImpl;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.deckfour.xes.model.impl.XAttributeMapImpl;
import org.deckfour.xes.model.impl.XEventImpl;
import org.deckfour.xes.model.impl.XLogImpl;
import org.deckfour.xes.model.impl.XTraceImpl;
import org.processmining.uncertaintymarker.help.Utilities;
import org.processmining.uncertaintymarker.models.Alternative;
import org.processmining.uncertaintymarker.models.Chunk;
import org.processmining.uncertaintymarker.parameters.Parameters;


/**The consolidation step is the first out of four steps performed by this plugin.
 * This class implements the consolidation process that is needed to transform the output of a classification algorithm into a Log with events.
 * Such a log can then be used for different types of process mining techniques. The input of this class contains probability distributions over
 * several prediction classes, which are produced by a classification algorithm in certain time windows. In this class the probabilities in this time windows
 * are aggregated, as long as the maximum prediction within the distributions is the same. When it changes for a certain number of time windows in a row, these
 * time windows will be summarized into a {@link Chunk}, with information about the label of the best prediction and the average probability/confidence of the summarized time windows.
 * 
 * @author Oliver
 */
public class Consolidation{

	//list of elements which aggregate time windows
	 List<Chunk> chunks = new ArrayList<>();
	
	//minimal number of time windows in a row that are neccessary to be not considered as outlier
	 int chunkCondition = -1; //TODO add as a user parameter?
	
	/**The main method of the consolidation class,invokes all submethods.
	 * 
	 * @param probLog - A classification output containing probability distributions. Class labels as column names, Probabilities with "."
	 * @param parameters 
	 * 
	 * @return A concise log containing traces, with single events in an order, which aggregate the probability information from the probLog
	 * */
	protected XLog preprocess(XLog probLog, Parameters parameters) {
		
		XLog conciseLog = new XLogImpl(probLog.getAttributes());
		chunkCondition = parameters.getChunkMinimum();
		
		for(XTrace trace: probLog){
			
			//add traces name attribute from probability log to concise log
			XTrace t = new XTraceImpl(trace.getAttributes());
			t.getAttributes().put("concept:name", trace.getAttributes().get("concept:name"));
			conciseLog.add(t);
			
			//create first chunk in trace
			Chunk chunk = new Chunk();
			chunks.add(chunk);
			
			int currentRow=0;
			
			//for each row
			for(XEvent row: trace){

				currentRow++;
				
				XAttributeMapImpl mapOfRow = (XAttributeMapImpl) row.getAttributes();
				
				ArrayList<XAttributeContinuousImpl> maxima = findMaxima(mapOfRow);
				
				//check for and handle equal maxima
				String maxLabel = checkForUnclearChunkDecision(chunk,maxima,currentRow);
				
				if(currentRow==1)
					chunk.setLabel(maxLabel);
				
				// same or new chunk?
				if(maxLabel.equals(chunk.getLabel())){
					chunk.addItem(mapOfRow);
				}
				else{
					chunk.computeAvg(); 
					
					//ADD NEW CHUNK
					XAttributeContinuousImpl nextLabel = findMaxima(mapOfRow).get(0);
					chunk = new Chunk(nextLabel.getKey());
					chunks.add(chunk);
					
					chunk.addItem(mapOfRow);
				}
			}
		
			if(parameters.getPruningAlgorithm().equals("Low-Frequency (Post-Pruning)"))
				removeOutliers(t);
			
			addChunksToConciseTrace(t);
			chunks = new ArrayList<>();
		}
		return conciseLog;
	}

	/**Checks whether the maxima of the current row are equal.
	 * If they are, it is not clear which event is the main prediction, but this is important
	 * for the alignment in the next step.
	 * 
	 * Then it is first checked whether one of the labels with equal maxima is already label of the current chunk.
	 * If it is, the label for this row is assumed to be the same. If not, it is saved in a list of candidates and in the next
	 * invocation of this method (which happens in the next row),the row label is compared to these candidates.
	 * 
	 * @param chunk - the current chunk that is collecting rows
	 * @param maxima - the rows maxima
	 * @param currentRow - counter showing which row is currently in processing
	 * @return the label the chunk should get
	 */
	private  String checkForUnclearChunkDecision(Chunk chunk, ArrayList<XAttributeContinuousImpl> maxima,int currentRow) {
		
		XAttributeContinuousImpl max = maxima.get(0);
		String maxLabel= max.getKey();
		
		//check if equal maxima and which label was before
		int firstUnclearRow = 0;
		int numberOfUnclearRows = 0;
		ArrayList<XAttributeContinuousImpl> candidates = new ArrayList<>();
		
		if( maxima.get(0).getValue()==maxima.get(1).getValue() && currentRow!=1 ){
			if(maxima.get(0).getKey().equals(chunk.getLabel()) || maxima.get(1).getKey().equals(chunk.getLabel())){
				maxLabel=chunk.getLabel();
			}
			else{
				candidates.add(maxima.get(0));
				candidates.add(maxima.get(1));
				if(numberOfUnclearRows==1)
					firstUnclearRow = currentRow;
				numberOfUnclearRows++;
			}
		}
		
		if(candidates.size()>0 && firstUnclearRow==currentRow-numberOfUnclearRows){
			for(XAttributeContinuousImpl c : candidates){
				if(maxima.get(0).equals(c.getKey()))
					maxLabel = maxima.get(0).getKey();
				if(maxima.get(1).equals(c.getKey()))
					maxLabel = maxima.get(1).getKey();
			}
			candidates.clear();
		}		
		return maxLabel;
	}

	/** Removes small chunks (i.e. with fewer rows inside than the chunk condition) from the chunks collection.
	 *  The purpose of this method is to exclude short-period missclassifications. 
	 */
	private void removeOutliers(XTrace t){

		for(int i=0;i<chunks.size()-1;i++){
			Chunk curr = chunks.get(i);
			if(curr.getRowsInChunk()<chunkCondition)
				chunks.remove(curr);
		}	
		for(int i2=0;i2<chunks.size()-1;i2++){
		
			Chunk curr2 = chunks.get(i2);
			Chunk next = chunks.get(i2+1);
	
			if(curr2.getLabel().equals(next.getLabel())){
				curr2.addAlternativesFrom(next);
				chunks.remove(next);
			}
		}
	}


	/**Creates events from the currently existing chunk collection and stores them into the input trace
	 * which is already added to the new concise log.
	 * 
	 * @param t - The input trace the event belongs to
	 */
	private void addChunksToConciseTrace(XTrace t){
		
		for(Chunk chunk : chunks){
			
			//CHUNK TO CONISE LOG
//			XAttributeMapImpl newMap = new XAttributeMapImpl();
			
//			XAttributeLiteralImpl name = new XAttributeLiteralImpl("concept:name", chunk.getLabel());
//			newMap.put("concept:name", name);
//			XAttributeContinuousImpl prob = new XAttributeContinuousImpl("prob", Utilities.round(chunk.getAvg(),3));
//			newMap.put("prob", prob);
//			XAttributeLiteralImpl altName = new XAttributeLiteralImpl("altName", chunk.getAllAlternatives().get(1).getLabel());
//			newMap.put("altName", altName);
//			XAttributeContinuousImpl altProb = new XAttributeContinuousImpl("altProb", Utilities.round(chunk.getAllAlternatives().get(1).getAvg(),3));
//			newMap.put("altProb", altProb);
//			
			XAttributeMapImpl newMap = new XAttributeMapImpl();
			for(Alternative a: chunk.getAllAlternatives()){
				String columnName = a.getLabel();
				double rowValue = a.getAvg();
			XAttributeContinuousImpl curr = new XAttributeContinuousImpl(columnName, Utilities.round(rowValue,3));
				newMap.put(columnName, curr);
			}
			
			XAttributeLiteralImpl name = new XAttributeLiteralImpl("concept:name", chunk.getLabel());
			newMap.put("concept:name", name);
			
			XEvent e = new XEventImpl(newMap);
			
			t.add(e);	
		}
	}
	
	/*** Finds the two best maxima in the attribute map from the input and returns a list containing XAttributes of it.
	 * @param mp - XAttributeMap containing entries, each entry is one row of the log/csv-File. 
	 * @return Key-Value-Pair containing the maximum in the row.
	 */
	public  ArrayList<XAttributeContinuousImpl> findMaxima(XAttributeMap mp) {
		
		ArrayList<XAttributeContinuousImpl> maxima = new ArrayList<>();
		
		XAttributeContinuousImpl max  = new XAttributeContinuousImpl("dummy", 0);
		XAttributeContinuousImpl nextMax  = new XAttributeContinuousImpl("dummy2", 0);
		XAttributeContinuousImpl tmp;
		
		for (Map.Entry<String, XAttribute> entry : mp.entrySet())
		{
		
			String key = entry.getKey();
			XAttribute val = entry.getValue();
			
	        if(val!=null && val instanceof XAttributeContinuousImpl){ //TODO check how to avoid failure when directly load XES-Log (val has other type)
	        	XAttributeContinuousImpl val2 = (XAttributeContinuousImpl)val;
	        	double value = val2.getValue();
	        	tmp = new XAttributeContinuousImpl(key,value);
	            if(tmp.getValue() > nextMax.getValue()){
	            	if(tmp.getValue()> max.getValue()){	            		
	            		nextMax = max;
	            		max= tmp;
	            	}
	            	else
	            		nextMax = tmp;
	            }
	        }
		}
		maxima.add(max);
		maxima.add(nextMax);
		
		return maxima;
	}
	
	/**Adds an occurrence number to each event, which is telling how often this event label occured in the enclosing trace already.	 */
	public void addingOccurrenceNumbers(XLog input){
		
		for(XTrace t: input){
			ArrayList<Map.Entry<String, Integer>> occurenceCounters = new ArrayList<>();
			for(XEvent e: t){
				
				XAttributeLiteralImpl labelAtt = (XAttributeLiteralImpl) e.getAttributes().get("concept:name");
				String label = labelAtt.getValue();
				
				//if already existent
				boolean exist = false;
				for(Map.Entry<String, Integer> entry : occurenceCounters){
					if(entry.getKey().equals(label)){
						exist=true;
						label += "" + entry.getValue();
						XAttributeLiteralImpl newLabel = new XAttributeLiteralImpl("concept:name", label);
						e.getAttributes().put("concept:name", newLabel);
						entry.setValue(entry.getValue()+1);
					}
				}
				
				if(!exist){
					Map.Entry<String, Integer> entry = new AbstractMap.SimpleEntry<String, Integer>(label,1);
					occurenceCounters.add(entry);
					label += "" + entry.getValue();
					XAttributeLiteralImpl newLabel = new XAttributeLiteralImpl("concept:name", label);
					e.getAttributes().put("concept:name", newLabel);
					entry.setValue(entry.getValue()+1);
				}
			}
		}
	}
}
