package org.processmining.uncertaintymarker.algorithms;

import java.util.ArrayList;

import javax.swing.JLabel;
import javax.swing.JPanel;

import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.connections.Connection;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetFactory;
import org.processmining.processtree.ProcessTree;
import org.processmining.uncertaintymarker.models.Alignment;
import org.processmining.uncertaintymarker.models.SuccessorMatrix;
import org.processmining.uncertaintymarker.parameters.Parameters;

/**Main Class of this plugins internal application logic.
 * It controls the flow of the mining process via invocation of the main methods of the other classes in this package.
 */
public class MainAlgorithm {

	XLog alignedLog = null;
	XLog preprocessedLog;
	Petrinet petriNet = null;
	Petrinet uncertaintyNet = null;
	JPanel panel;
	
	/**Implements the main method of the plugin, 
	 * controls the flow of the whole mining process by starting the different steps (in dependency from the parameters).
	 * 
	 * @param context - The context in which the plugin runs.
	 * @param probLog - A log containing different event predictions as probability distributions, generated by a classification algorithm 
	 * @param parameters - The parameters used in different parts of the plugin and provided by the user through the UI.
	 * 
	 * @return The mined petri net with probability annotations, the concise log produced by the time window consolidation and the aligned version of this log
	 */
	public Object[] apply(PluginContext context, XLog probLog, Parameters parameters) {

		SuccessorMatrix sm = new SuccessorMatrix(probLog);
		
		//Step 0: HMM
		if(parameters.isChunkingActive() && parameters.getPruningAlgorithm().equals("HMM (Pre-Pruning)")){
			HmmPruning hmm = new HmmPruning(sm, probLog);
			probLog = hmm.outputLog;
			XLog hmmLog = hmm.outputLog;
			
			  context.getProvidedObjectManager().createProvidedObject(
                      "HMM-pruned log " ,
                      hmmLog, XLog.class, context);
		}
		
		//Step 1: Time Window Consolidation
		if(parameters.isChunkingActive()){
			Consolidation c = new Consolidation();
			preprocessedLog = c.preprocess(probLog,parameters);
//			c.addingOccurrenceNumbers(preprocessedLog);
		}
		else{
			preprocessedLog = probLog;
		}

		
		//Step 2: Trace Alignment
		if(preprocessedLog.size()>1){

			//transfering parameters
			int[] alignmentParameter = new int[8];
			alignmentParameter[0] = parameters.getIndelScore();
			alignmentParameter[1] = parameters.getSubScore1();
			alignmentParameter[2] = parameters.getSubScore2();
			alignmentParameter[3] = parameters.getSubScore3();
			alignmentParameter[4] = parameters.getEvalScore1();
			alignmentParameter[5] = parameters.getEvalScore2();
			alignmentParameter[6] = parameters.getEvalScore3();
			alignmentParameter[7] = parameters.getEvalScore4();
			
			ArrayList<Alignment> startTraces = new ArrayList<>();
			
			for(XTrace t: preprocessedLog){
				String tLabel = t.getAttributes().get("concept:name").toString();
				ArrayList<XTrace> l = new ArrayList<>();
				l.add(t);
				startTraces.add(new Alignment(tLabel,null, null, l));
			}
			
			TraceAlignment ta = new TraceAlignment();
			ta.transferParameters(alignmentParameter);
			Object[] tmp = ta.recursiveTA(startTraces,0);
			alignedLog = (XLog) tmp[0];
			
			panel = new JPanel();
			panel.add(new JLabel((String)tmp[1]));
		}
		else{
			alignedLog = preprocessedLog;
		}
		
		//Step 3: Process Discovery & Trace Consolidation
		String algorithm = parameters.getAlgorithm();
		
		switch (algorithm) {
			case "alpha" :
//			Object[] objs = AlphaMinerPlugin.applyAlphaClassic(context, preprocessedLog, preprocessedLog.getClassifiers().get(0));
//			net =  (Petrinet) objs[0]; 
			try {
				petriNet = context.tryToFindOrConstructFirstNamedObject(Petrinet.class,"Alpha Miner", Connection.class, "", preprocessedLog);
			} catch (ConnectionCannotBeObtained e1) {
				e1.printStackTrace();
			}
			break;
			
			case "inductive" :
				try {
					petriNet = context.tryToFindOrConstructFirstNamedObject(Petrinet.class,"Mine Petri net with Inductive Miner", Connection.class, "", preprocessedLog);
				} catch (ConnectionCannotBeObtained e1) {
					e1.printStackTrace();
				}
			break;
				
			case "etm" :
				try {
					ProcessTree tree = context.tryToFindOrConstructFirstNamedObject(ProcessTree.class,"Mine a Process Tree with ETMd", Connection.class, "", preprocessedLog);
					petriNet = context.tryToFindOrConstructFirstNamedObject(Petrinet.class,"Convert Process Tree to Petri Net", Connection.class, "", tree);
				} catch (ConnectionCannotBeObtained e) {
					e.printStackTrace();
				}
			break;
				
			case "alignedStepsDiscovery" :
				double th = parameters.getAsdInclusionThreshold();
				AlignedStepDiscovery asd = new AlignedStepDiscovery();
				petriNet = asd.discoverPetrinet(alignedLog,th,context);
			break;

			default :
			break;
		}
		
		//clone Petri net for output, before it is enhanced
		uncertaintyNet = PetrinetFactory.clonePetrinet(petriNet);
		
		//Step 4: Probability Annotation to Petrinet
		NetAttachment na = new NetAttachment();
		na.attachInformation(uncertaintyNet, context, alignedLog,parameters);
		
				
		//RETURN NET, CONCISE AND ALIGNED LOG
		Object[] output = new Object[5];
		output[0] = uncertaintyNet;
		output[1] = petriNet;
		output[2] = preprocessedLog;
		output[3] = alignedLog;
		output[4] = panel;
		
		return output; 
	}
}