package org.processmining.uncertaintymarker.algorithms;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Iterator;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.deckfour.xes.model.impl.XAttributeMapImpl;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.connections.GraphLayoutConnection;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.petrinet.PetriNetVisualization;
import org.processmining.uncertaintymarker.help.Utilities;
import org.processmining.uncertaintymarker.models.Alternative;
import org.processmining.uncertaintymarker.models.TransitionData;
import org.processmining.uncertaintymarker.parameters.Parameters;

/**This class implements the fourth and last step of the plugin. After a Petrinet is mined by an arbitrary discovery algorithm (e.g. the {@link AlignedStepDiscovery}),
 * it is used to attach the alternative predictions and to add the probabilities. Whether an alternative is included 
 * or not depends on a probability difference of the (average) confidence of the first prediction and the (average) confidence of its alternative.
 * The accepted difference can be defined by the user in the parameter setup.
 * 
 * @author Oliver
 *
 */
public class NetAttachment{
	
	 Petrinet newNet = null;
	
	 XLog log = null;
	 int logLength;
	 int logDepth;
	
	 PluginContext context;
	
	 Parameters parameters;
	
	 ArrayList<Transition> bestClassTransitions = new ArrayList<Transition>();

	 ArrayList<ArrayList<TransitionData>> transitionsByColumn = new ArrayList<>();
	
	 ArrayList<TransitionData> transitionsWithSufficientFrequency = new ArrayList<>();
	
	public  Petrinet attachInformation(Petrinet net, PluginContext context,XLog alignedLog,Parameters parameters ){
		
		newNet = net;
		log = alignedLog;
		this.context = context;
		this.parameters = parameters;
		
//		removeOccurenceNumbers(log,newNet);
		
		aggregateByColumn();

		addIDsToNet();
		
		addAlternativesWithProbabilities();
		
		double val= parameters.getAsdInclusionThreshold() ;
//		newNet.addTransition( "<html> " + String.valueOf(val) + "</html>");
//		newNet.addTransition("<html>Parameters: <br><br> algorithm:&nbsp;"+ parameters.getAlgorithm() + " <br> inclusion&nbsp;threshold:&nbsp;" + String.valueOf(val) + " <br> difference&nbsp;threshold:&nbsp;" + parameters.getDifferenceThreshold() + "</html>");
		newNet.addTransition("<html> Parameters: <br><br>" + parameters.getAlgorithm() +"<br>"+ String.valueOf(val) + "<br>" + parameters.getDifferenceThreshold() + "</html>");

		
		//connect and visualize before position can be changed
		GraphLayoutConnection layout = new GraphLayoutConnection(newNet);
        context.addConnection(layout);	
		PetriNetVisualization v = new PetriNetVisualization();
		v.visualize(context, newNet);
		
		changeAlternativesPositions(layout);
		
		addProbabilitiesToEvents();
		
		removeIDsFromNet();
		
		
		return newNet;
	}

	/**Adds IDs to all transition labels in the petri net, which are later used to map them to their alternative transitions .
	 * Add those transitions to the bestClassTransitions-list.
	 * After mapping, the IDs are removed again from all the transition labels by the removeIds-method.
	 */
	private  void addIDsToNet() {
		
		//to avoid problems with remaining stuff from  cache / previous invocation of the plugin
		bestClassTransitions.clear();
		
		//get all transitions in NET and add IDs
		Iterator<Transition> itTransitions = newNet.getTransitions().iterator();
		int index = 0;
		while(itTransitions.hasNext()){
			//copy from iterator to list (because iterator must not be changed while being accessed)
			Transition t = itTransitions.next();
			bestClassTransitions.add(t);
			
			//adding id to label
			String labelWithId = index + " " + t.getLabel();
			t.getAttributeMap().put(AttributeMap.LABEL, labelWithId);
			index++;
			
			//change width of new t
			Dimension size2 = (Dimension) t.getAttributeMap().get(AttributeMap.SIZE);
			size2.setSize(100, size2.getHeight());
			t.getAttributeMap().put(AttributeMap.SIZE,size2);
		}
	}


/**Fill the list transitionsByColumn with lists of event representations of type {@link TransitionData}. 
 * Each of this lists represents one column in the aligned log. The occurrences of an event in a particular column are counted
 * and the averages of the event and its alternatives are computed in the addItem-method. 
 */
	private  void aggregateByColumn() {

		logLength = log.get(0).size();
		logDepth = log.size();
		
		transitionsByColumn.clear();
		transitionsWithSufficientFrequency.clear();
		
		for(int column=0;column< logLength;column++){
			ArrayList<TransitionData> transitions= new ArrayList<>();
			//fill list
			for(XTrace t: log){
				
				XAttributeMapImpl map = new XAttributeMapImpl(t.get(column).getAttributes());
	
				XAttributeLiteralImpl xAtt = (XAttributeLiteralImpl) map.get("concept:name");
				String currEvent = xAtt.getValue(); 
				
				boolean flag=true;
				for(int i=0;i<transitions.size();i++){ //add row information to existing element
					TransitionData tr=transitions.get(i);
					if(tr.getLabel().equals(currEvent)){
						tr.addItem(map,column);
						flag =false;
					}
				}
				if(flag ){ //add new element
					TransitionData newTr2 = new TransitionData();
					newTr2.addItem(map,column);
					transitions.add(newTr2);
				}
			}
			
			transitionsByColumn.add(transitions);
		}
	}

	/**Removes ids from transition labels. They were only used for mapping of event transitions and their alternative transitions,
	 * are not needed anymore afterwards and might confuse the user.
	 */
	private  void removeIDsFromNet() {

		Iterator<Transition> it = newNet.getTransitions().iterator();
		
		while(it.hasNext()){
			Transition curr = it.next();	

			//remove id from label
			String labelWithId = curr.getLabel();
			String[] labelParts = curr.getLabel().split("\\(");
			
			String cleanPart1 = labelParts[0].replaceAll("[0-9]+ ", "").replaceAll("\\s+","");
			String cleanLabel = cleanPart1;
			
			if(labelParts.length>1){
				cleanLabel += " (" +labelParts[1];
			}
			
			curr.getAttributeMap().put(AttributeMap.LABEL, cleanLabel);
		}
	}

	/**Adds alternatives to the petri net generated by the third step, their probabilities are included in the label.
 	 * As it is not possible to modify the Transition-class, the id needs to be added to the label. 
 	 * Then, for each list (representing a column) in the
	 * transitionByColumn-list , each alternative in each aggregated event in this column is checked against the 'Alternative Probability Difference Threshold'.
	 * Only if it the check is successful, the alternative is included. If the discovery algorithm used before was AlignedStepDiscovery, it is first matched against the 'Columnshare Discovery Threshold',
	 * to ensure a correct mapping. 
	 */
	private  void addAlternativesWithProbabilities(){

		transitionsWithSufficientFrequency.clear();
		
		//to avoid mispositioning of alternatives in net
		ArrayList<Transition> alreadyEdited = new ArrayList<>();
		Transition tmp = null;

		//each event in each column of the alignment
		for(ArrayList<TransitionData> column : transitionsByColumn){
			for(TransitionData e : column){
			
				//exclude column if consists only of gap-signs
				if(e.getLabel().equals("-") && column.size()==1)
					continue;
				
				double relativeFrequency = ((double)e.getFrequency())/logDepth;
				if(relativeFrequency >= parameters.getAsdInclusionThreshold() || !parameters.getAlgorithm().equals("alignedStepsDiscovery")){
					transitionsWithSufficientFrequency.add(e);

					ArrayList<Alternative> alternatives = e.getAllAlternatives();
					for(int i=1;i<alternatives.size();i++){
						Alternative a = alternatives.get(i);
						
						//get match and add id to new transition
						double difference = e.getAvg() - a.getAvg();
					
						if(difference <= parameters.getDifferenceThreshold() && Utilities.round(a.getAvg(), 2)!= 0){
							
							// add all alternatives meeting thresholds
							for(int x=0;x<bestClassTransitions.size();x++){ 
								Transition t = bestClassTransitions.get(x); 
								if(!alreadyEdited.contains(t)){
									int id = getMatchId(t);
									String s = String.valueOf(id);
									int digits = s.length();
									String label_unchanged = t.getLabel();
									String label = t.getLabel().substring(digits+1).replaceAll(" \\([0-9].[0-9]+\\)", ""); //replaceAll("[0-9]+ ","")
									if(label.equals(e.getLabel())){
										
										String altLabel = id+" "+ a.getLabel() + " (" + String.format("%.2f", a.getAvg()) + ")";
										Transition alternativesTransition = newNet.addTransition(altLabel);
										
										//change width of new t
										Dimension size2 = (Dimension) alternativesTransition.getAttributeMap().get(AttributeMap.SIZE);
										size2.setSize(100, size2.getHeight());
										alternativesTransition.getAttributeMap().put(AttributeMap.SIZE,size2);
										tmp = t;
										break;
									}
								}
							}
						}
						if(i==alternatives.size()-1)
							alreadyEdited.add(tmp);
					}
				}
			}
		}
	}
	
	/**Returns the id from the label of the Transition t.
	 * This id is used for mapping of an event transition to its alternative transition. 
	 */
	private  int getMatchId(Transition t) {
		String[] ids1 = t.getLabel().split(" ");
		String ids2 = ids1[0];
		int id = Integer.parseInt(ids2);
		return id;
	}


	 /** Sets the position of the alternatives beneath the corresponding events.
	 *After the alternative transitions are added to the net by the addAlternatives-method,
	 * they are not yet visually mapped to their event transitions by positioning, but only by an id in the label. 
	 * Before this method can be applied, the net needs to be visualized using the {@link PetrinetVisualization} class.
	 */
	private  void changeAlternativesPositions(GraphLayoutConnection layout) {

		ArrayList<Transition> bestClasses = new ArrayList<Transition>();

		Iterator<Transition> it = newNet.getTransitions().iterator();
		//find best classes by using the brackets
		while(it.hasNext()){
			Transition curr = it.next();	
			
			if(!curr.getLabel().contains("(") && !(curr.getLabel().contains("Parameters:"))){
				bestClasses.add(curr);
			}
		}
		
		Iterator<Transition> it2 = newNet.getTransitions().iterator();
		int[] positionFactor = new int[bestClasses.size()];

		int prevHeight = 0;
		Point2D prevPos = null;
		
		while(it2.hasNext()){
			
			Transition curr = it2.next();
			
			

			//parameter display dummy transition
			if(curr.getLabel().contains("Parameters")){
				Point2D posSub =  new Point(0,(int)prevPos.getY()-200);
		        layout.setPosition(curr, posSub);
		        
				//change width of new t
				Dimension size2 = (Dimension) curr.getAttributeMap().get(AttributeMap.SIZE);
				size2.setSize(170, 70);
				curr.getAttributeMap().put(AttributeMap.SIZE,size2);
		        continue;
			}
				
				
			int index = 0;
			for(Transition bestClass: bestClasses){
				
				if(bestClass.getLabel().startsWith("<html>") || curr.getLabel().startsWith("<html>")){
					
					//change width of new t
					Dimension size2 = (Dimension) curr.getAttributeMap().get(AttributeMap.SIZE);
					size2.setSize(200, size2.getHeight());
					curr.getAttributeMap().put(AttributeMap.SIZE,size2);
					continue;
				}
				
				if(getMatchId(curr) == getMatchId(bestClass)){
					
					Point2D pos = layout.getPosition(bestClass);
					
					int x = (int) (pos.getX());
					int y = (int) (pos.getY() + 40*positionFactor[index]);
					
					Point2D posSub =  new Point(x,y);
					
			        layout.setPosition(curr, posSub);
			        
			        positionFactor[index]++;
				}
				
				index++;
			}
			Dimension d = (Dimension)curr.getAttributeMap().get(AttributeMap.SIZE);
			prevHeight = (int) d.getHeight();
			
			prevPos = layout.getPosition(curr);
		}
	}


	/** Adds the average probabilities of each event (with sufficient frequency) to its transition label in the net.
	 *  Invoked after alternatives are added, since the only difference between an event transition and an alternative transition in the net can be the label.
	 *  The addAlternatives()-method uses the fact that the event transition labels (in contrast to the alternative transitions) don't have brackets/probabilities yet.
	 *  Afterwards this method adds the missing probabilities
	 */
	private  void addProbabilitiesToEvents(){
		
		//get all transitions in NET
		Iterator<Transition> itTransitions = newNet.getTransitions().iterator();
		//iterator content must not change while iterating
		ArrayList<Transition> transitionsBefore = new ArrayList<Transition>();
		
		while(itTransitions.hasNext()){

			//copy from iterator to list (because iterator must not be changed while bein accessed)
			Transition t = itTransitions.next();
			if(t.getLabel().contains("("))
				continue;
			transitionsBefore.add(t);	
			
			String tLabel = (String) t.getAttributeMap().get(AttributeMap.LABEL);
			tLabel = tLabel.replaceAll("[0-9]+ ", "");
			for(int i=0;i<transitionsWithSufficientFrequency.size();i++){
				TransitionData td = transitionsWithSufficientFrequency.get(i);
				if(tLabel.equals(td.getLabel())){
					String labelWithProb = td.getLabel() + " (" + String.format("%.2f", td.getAvg()) + ")";
					t.getAttributeMap().put(AttributeMap.LABEL, labelWithProb);
					transitionsWithSufficientFrequency.remove(i);
					break;
				}
			}
		}
	}
	
	/**Removes occurrence numbers from transition labels and from event labels in the log, since they are not needed anymore after the trace alignment step,
	 * and would make the id-mapping more complex.
	 */
	protected static void removeOccurenceNumbers(XLog log,Petrinet net) {
		
		//remove occurenceCount from labels in log
		for(XTrace t:log){
			for(XEvent e:t){
				XAttributeLiteralImpl xAtt = (XAttributeLiteralImpl) e.getAttributes().get("concept:name");
				String label = xAtt.getValue();
				String newLabel = label.replaceAll("[0-9]+", "");
				xAtt.setValue(newLabel);
				e.getAttributes().put("concept:name", xAtt);
				
			}
		}

		//remove occurenceCount from labels in net
		Iterator<Transition> it = net.getTransitions().iterator();
		while(it.hasNext()){
				Transition curr = it.next();	
				String labelWithOccurence = curr.getLabel();
				String labelWithOutOccurence = labelWithOccurence.replaceAll("[0-9]+", "");
				curr.getAttributeMap().put(AttributeMap.LABEL, labelWithOutOccurence);
			}
	}
}
