package org.processmining.uncertaintymarker.algorithms;

import java.util.ArrayList;
import java.util.Collections;

import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.deckfour.xes.model.impl.XAttributeMapImpl;
import org.deckfour.xes.model.impl.XEventImpl;
import org.deckfour.xes.model.impl.XLogImpl;
import org.deckfour.xes.model.impl.XTraceImpl;
import org.processmining.uncertaintymarker.models.Alignment;
import org.processmining.uncertaintymarker.models.Scoringtable;

/** <p>The trace alignment step is the second out of four steps performed by this plugin. 
 * It is executed after {@link Consolidation} and before an arbitrary process discovery algorithm, e.g. {@link AlignedStepDiscovery}.
 * In this class {@link Alignment}s are generated by comparing traces with each other and adding gap-signs to them,
 * so that they have an equal length and are more similar in eventwise comparison of the traces. </p>
 * 
 * <p>First, traces are compared pairwise, generating alignments of size two.
 * Afterwards they are evaluated by a scoring function and the best alignments are aligned again, evaluated and selected for the next level of hierarchy.
 * This procedure continues until all traces are part of the final alignment. While the subscore- and indelscore-parameters are part of the scoring table building algorithm,
 * which is used to actually do an alignment, the evalscore-parameter determine how it is rated afterwards in order to select the best one. 
 * All these parameters can be set by the user.</p>
 * 
 * I reimplemented the algorithm from the (at development date) current version of
 * TraceAlignment in ProM from J. C. Boose, because it was not executable.
 * 
 * @author Oliver
 */
public class TraceAlignment {

	 String gapSign = "-";
	
	//NOT insert a gap sign with these constants in formula of building-step rate
	 int subScoreCase1 = 1; //A A
	 int subScoreCase2 = -1; //A B
	 int subScoreCase3 = -1; //A - , - A, --
	
	// insert gap sign with this constant in formula of building-step rate
	 int indelScoreCase1 = -1; // in all cases
	
	//rate result
	 int evalScoreCase1 =  10; // A A
	 int evalScoreCase2 =  -10; // A B
	 int evalScoreCase3 =  -1; // A B, ? - 
	 int evalScoreCase4 =  -1; // - A
	
	//hierarchy lvl
	 int lvl;
	
	/**In this constructor the scoring parameters are transfered from the user input to the objects datafields,
	 * in order to make the alignment configurable.
	 */
	public  void transferParameters(int[] alignmentParameter) {
			indelScoreCase1 = alignmentParameter[0];
			subScoreCase1	 = alignmentParameter[1];
			subScoreCase2 = alignmentParameter[2];
			subScoreCase3  = alignmentParameter[3];
			evalScoreCase1 = alignmentParameter[4];
			evalScoreCase2 = alignmentParameter[5];
			evalScoreCase3 = alignmentParameter[6];
			evalScoreCase4 = alignmentParameter[7];
	   }

	/**The recursive main method of this class, which controls the different substeps by invocation of the submethods.
	 * In the first invocation it starts the pairwise alignment of all the traces in the list by calling a certain method.
	 * Afterwards all potential pairs are evaluated with the scoring function and the best one are selected into the list parameter for the next round
	 * of self-invocation.
	 * In the next invocations it takes the resulting alignments from the previous one and starts the alignment of these with another method.
	 * It it stops when the list which contains the result alignments from the previous iteration only contains one alignment of all traces,
	 * and return this aligned log.
	 * 
	 * @param list - The resulting alignments of the current invocation round (count)
	 * @param count - The n-th invocation of the recursive method
	 * 
	 * @return - The aligned log
	 */
	public  Object[] recursiveTA(ArrayList<Alignment> list,int count){
		
		lvl = count;
		
		//termination criterion
		if(list.size()==1){
			System.out.println();
			System.out.println("FINAL ALIGNMENT: ");
			System.out.println();
			String printable = printAlignment(list.get(0));
			
			XLog alignedLog = new XLogImpl(new XAttributeMapImpl());			
			for(XTrace t: list.get(0).getAlignedTraces()){
				alignedLog.add(t);
			}
			Object[] result = new Object[2];
			result[0]= alignedLog;
			result[1]= printable;
			return result;
		}
		
        //creates all possible pairwise alignments and rates them
		ArrayList<Alignment> possibleAlignments = new ArrayList<>();
		ArrayList<Integer> evaluation = new ArrayList<>();	
		
		for(int i=0;i<list.size();i++){
			for(int j=i+1;j<list.size();j++){
				Alignment a1 = list.get(i);
				Alignment a2 = list.get(j);
				Alignment newAlignment = null;
				
				if(a1.getPredecessor_2()==null && a2.getPredecessor_2()==null ){
					newAlignment = pairWiseTraceAlignment(a1.getAlignedTraces().get(0), a2.getAlignedTraces().get(0));
				}
				else{
					newAlignment = alignAlignments(a1, a2);
				}

				//Collect all possible alignment groups
				possibleAlignments.add(newAlignment);
				
				//evaluate each of these groups
				int sumOfPairs = 0;
				for(XTrace t1: newAlignment.getAlignedTraces()){
					for(XTrace t2: newAlignment.getAlignedTraces()){
						if(t1==t2) continue;
						sumOfPairs += evaluatePartOfAlignment(t1,t2);
					}
				}
				evaluation.add(sumOfPairs);
			}
		}
		//collect best alignments at current level by using evaluations
		ArrayList<Alignment> newList = findBestAlignsAtCurrentLevel(possibleAlignments, evaluation, list,count);

		
		return recursiveTA(newList,count+1);
	}

	/**<p>Aligns two traces by adding gap-signs to them.</p>
	 * 
	 * <p>First a scoring table (2D-Integer matrix) is build.
	 * Each cells value is composed of the value of one of their neighbors(top,diagonal-top-left,left) + a certain function including the users scoring parameters,
	 * depending on which value is maximal. </p>
	 * 
	 * <p>Afterwards the algorithm start at the bottom-right end of the table and checks which of the aforementioned neighbor values is maximal.
	 * Depending on this one of three actions is performed: Adding a gap-sign to T1, not adding any gap-sign, or adding one to T2.
	 * Then it jumps to this maximal neighbor-value and does the same checking. </p>
	 * 
	 * <p>The pairwise alignment is done after the algorithm ended in the top-left corner of the table.</p>
	 */
	public  Alignment pairWiseTraceAlignment(XTrace t1, XTrace t2){
				
				//setup scoring matrix
//				int[] [] scoringTable = new int[t1.size()+1] [t2.size()+1];
				Scoringtable st = new Scoringtable(t1.size()+1,t2.size()+1);
				st.scoringTable[0][0] = 0;
				st.scoringTable[1][0] = 0;
				st.scoringTable[0][1] = 0;
				
				// fill first row
				for(int i=2;i<=t1.size();i++){
					st.scoringTable[i][0] = st.scoringTable[i-1][0] +  indelScore(t1,i-1,t1,i-2);
				}
				// fill first column
				for(int j=2;j<=t2.size();j++){
					st.scoringTable[0][j] = st.scoringTable[0][j-1] +  indelScore(t2,j-1,t2,j-2);
				}
				
				//compute matrix
				for(int i=1;i<=t1.size();i++){
					for(int j=1; j<=t2.size();j++){
						
						int scoreOption1 = st.scoringTable[i-1][j-1] + substitutionScore(t1, i, t2, j)  ;
						int scoreOption2 = st.scoringTable[i-1][j] + indelScoreCase1; //indelScore(t1, i, t1, i-1, gapSign); TODO check for change
						int scoreOption3 = st.scoringTable[i][j-1] + indelScoreCase1; //indelScore(t1, i, t1, i-1, gapSign);
						
						int max = Math.max(scoreOption1, scoreOption2);
						max = Math.max(max, scoreOption3);
						
						st.scoringTable[i][j]= max;
					}
				}
				
				//compute alignment from matrix
				XTrace alignedTraceT1 = new XTraceImpl(t1.getAttributes());
				XTrace alignedTraceT2 = new XTraceImpl(t2.getAttributes());
				ArrayList<XEvent> eventsT1 = new ArrayList<>();
				ArrayList<XEvent> eventsT2 = new ArrayList<>();
				
				int i= t1.size(); int j= t2.size();
		
				while(i>0 || j>0){
					
					int max = -10000000;
					int left = -10000000;
					int top = -10000000;
					int diagonal = -10000000;
					//get direction
					if(i==0){ //first row reached
						left = st.scoringTable[i][j-1];
						max = left;
					}
					else if(j==0){ //first column reached
						top = st.scoringTable[i-1][j];
						max= top;
					}
					else{
						left = st.scoringTable[i][j-1];
						top = st.scoringTable[i-1][j];
						diagonal = st.scoringTable[i-1][j-1];
						
						max = Math.max(left, top);
						max = Math.max(max, diagonal);
					}
					
					//create alignedEvents & move to new direction
					XEvent alignedEventT1 = new XEventImpl();
					XEvent alignedEventT2 = new XEventImpl();
					
					if(max==diagonal){
						alignedEventT1.setAttributes(t1.get(i-1).getAttributes()); //create
						alignedEventT2.setAttributes(t2.get(j-1).getAttributes());
						
						i -= 1; //move
						j -= 1;
					} 
					else if(max==top){
						alignedEventT1.setAttributes(t1.get(i-1).getAttributes()); //create
						XAttributeMap m = new XAttributeMapImpl();
						XAttributeLiteralImpl l = new XAttributeLiteralImpl("concept:name", gapSign);
						m.put("concept:name", l);
						alignedEventT2.setAttributes(m);
						
						i -= 1; //move
					}
					else{
						alignedEventT2.setAttributes(t2.get(j-1).getAttributes()); //create
						XAttributeMap m = new XAttributeMapImpl();
						XAttributeLiteralImpl l = new XAttributeLiteralImpl("concept:name", gapSign);
						m.put("concept:name", l);
						alignedEventT1.setAttributes(m);
						
						j -= 1; //move
					}
					eventsT1.add(alignedEventT1);
					eventsT2.add(alignedEventT2);
				}
				
				for(int index = eventsT1.size()-1;index>=0;index--){
					alignedTraceT1.add(eventsT1.get(index));
					alignedTraceT2.add(eventsT2.get(index));
				}

				
				
				ArrayList<XTrace> alignedTraces = new ArrayList<>();
				alignedTraces.add(alignedTraceT1);
				alignedTraces.add(alignedTraceT2);

				String labelT1 = ((XAttributeLiteralImpl) t1.getAttributes().get("concept:name")).toString();
				String labelT2 = ((XAttributeLiteralImpl) t2.getAttributes().get("concept:name")).toString();
				
				Alignment alignment = new Alignment("A 0",labelT1, labelT2, alignedTraces);
				
				return alignment;
}

	/**<p>A function that is used in the construction of the scoring table in the pairwiseTraceAlignment-method.
	 * It influences how the alignment is build, by contributing to the decision whether it would be advantageous for the score to NOT add a gap sign in any of the traces.</p>
	 * 
	 * <p>The event label at pos1 in traceVal1 is compared to the event label at pos2 in traceVal2.
	 * It should be mentioned that in some cases traceVal1 and traceVal2 is the same trace.</p>
	 * 
	 * 
	 * @param traceVal1 - A trace that should be investigated at pos1 in relation to trace2 at pos2.
	 * @param gapSign - The string/letter that should be used as a gap sign
	 * 
	 * @return Integer indicating whether not adding a gap-sign is advantageous with this events or not. (very high -> not add,very low/negative -> add)
	 */
	private  int substitutionScore( XTrace traceVal1, int pos1, XTrace traceVal2,  int pos2){
		
		XEvent elemT1 = traceVal1.get(pos1-1);
		XEvent elemT2 = traceVal2.get(pos2-1);
		
		XAttributeLiteralImpl elemAttT1 = (XAttributeLiteralImpl) elemT1.getAttributes().get("concept:name");
		String elemLabelT1 = elemAttT1.getValue();
		XAttributeLiteralImpl elemAttT2 = (XAttributeLiteralImpl) elemT2.getAttributes().get("concept:name");
		String elemLabelT2 = elemAttT2.getValue();
		
		if(!elemLabelT1.equals(gapSign) && !elemLabelT2.equals(gapSign)){
			if(elemLabelT1.equals(elemLabelT2)){
				return subScoreCase1; //A A (event labels are the same)
			}
				return subScoreCase2; //A B (labels are different)
		}
		return subScoreCase3; // A - , B - , - -
	}
	
	/**A function that is used in the construction of the scoring table in the pairwiseTraceAlignment-method.
	 * It influences how the alignment is build, by contributing to the decision whether it would be advantageous for the score to NOT add a gap sign in any of the traces.
	 */	
	private  int indelScore( XTrace traceVal1, int pos1, XTrace traceVal2,  int pos2){
	
//		//current events at i in T1 & j in T2
//		XAttributeLiteralImpl elemAttT1 = (XAttributeLiteralImpl) traceVal1.get(pos1).getAttributes().get("concept:name");
//		String elemLabelT1 = elemAttT1.getValue();
//		XAttributeLiteralImpl elemAttT2 = (XAttributeLiteralImpl) traceVal2.get(pos2).getAttributes().get("concept:name");
//		String elemLabelT2 = elemAttT2.getValue();
//		
//		//previous events at i-1 in T1 & j-1 in T2
//		XAttributeLiteralImpl prevElemAttT1 = (XAttributeLiteralImpl) traceVal1.get(pos1-1).getAttributes().get("concept:name");
//		String prevElemLabelT1 = prevElemAttT1.getValue();
//		XAttributeLiteralImpl prevElemAttT2 = (XAttributeLiteralImpl) traceVal2.get(pos1-1).getAttributes().get("concept:name");
//		String prevElemLabelT2 = prevElemAttT2.getValue();
		
		return indelScoreCase1;
//		if( (!elemLabelT1.equals(gapSign) && 
//				 !prevElemLabelT1.equals(gapSign) && 
//				  elemLabelT2.equals(gapSign) 
//				 ) ||
//				 (!elemLabelT2.equals(gapSign) && 
//				  !prevElemLabelT2.equals(gapSign) && 
//				  elemLabelT1.equals(gapSign) 
//				 )	){
//				
//			//indelscore -1
//			}
	}
	
	/**<p>Aligns two alignments by adding gap-signs to their traces.</p>
	 * 
	 * This method is quite similar to the pairwiseTraceAlignment-method, but works with {@link Alignment}s instead of single traces as inputs.
	 * For this purpose it uses its own submethods for scoring.
	 * 
	 * <p>First a scoring table (2D-Integer matrix) is build.
	 * Each cells value is composed of the value of one of their neighbors(top,diagonal-top-left,left) + a certain function including the users scoring parameters,
	 * depending on which value is maximal. </p>
	 * 
	 * <p>Afterwards the algorithm start at the bottom-right end of the table and checks which of the aforementioned neighbor values is maximal.
	 * Depending on this one of three actions is performed: Adding a gap-sign to all traces in A1, not adding any gap-sign, or adding a gap-sign to all traces in A2.
	 * Then it jumps to this maximal neighbor-value and does the same checking. </p>
	 * 
	 * <p>The alignment of the input sub-alignments is done after the algorithm ended in the top-left corner of the table.</p>
	 */
	public  Alignment alignAlignments(Alignment a1, Alignment a2){
	
		int dimA1 = a1.getAlignedTraces().get(0).size();	
		int dimA2 = a2.getAlignedTraces().get(0).size();	
		
		//setup scoring matrix
		Scoringtable st = new Scoringtable(dimA1+1,dimA2+1);
//		int[] [] scoringTable = new int[dimA1+1] [dimA2+1];
		st.scoringTable[0][0] = 0;
		st.scoringTable[1][0] = 0;
		st.scoringTable[0][1] = 0;
		
		// fill first row
		for(int i=1;i<=dimA1;i++){
			st.scoringTable[i][0] = st.scoringTable[i-1][0] + indelScoreBetweenAlignments(a1, i );
		}
		// fill first column   //TODO check indelScore or delete it
		for(int j=1;j<=dimA2;j++){
			st.scoringTable[0][j] = st.scoringTable[0][j-1] + indelScoreBetweenAlignments(a2, j );
		}
		
		//compute matrix
		for(int i=1;i<=dimA1;i++){
			for(int j=1; j<=dimA2;j++){
				//TODO change functions
				int scoreOption1 = st.scoringTable[i-1][j-1] + subScoreBetweenAlignments(a1, i, a2, j);
				int scoreOption2 = st.scoringTable[i-1][j] + indelScoreBetweenAlignments(a1, i );
				int scoreOption3 = st.scoringTable[i][j-1] + indelScoreBetweenAlignments(a2,j);
				
				int max = Math.max(scoreOption1, scoreOption2);
				max = Math.max(max, scoreOption3);
				
				st.scoringTable[i][j]= max;
			}
		}
		
		ArrayList<XTrace> alignedAlignmentA1Reverse = new ArrayList<>();
		ArrayList<XTrace> alignedAlignmentA2Reverse = new ArrayList<>();
		
		//create new traces from aligned ones of previous round
		for(XTrace t: a1.getAlignedTraces()){
			XTrace alignedTraceA1 = new XTraceImpl(t.getAttributes()); 
			alignedAlignmentA1Reverse.add(alignedTraceA1);
		}
		for(XTrace t: a2.getAlignedTraces()){
			XTrace alignedTraceA2 = new XTraceImpl(t.getAttributes()); 
			alignedAlignmentA2Reverse.add(alignedTraceA2);
		}
		
		
		int i= dimA1; int j= dimA2;
		int count = 0;
		
		//move to new direction
		while(i>0 || j>0){
			
			int max =      -10000000;
			int left =     -10000000;
			int top =      -10000000;
			int diagonal = -10000000;
			
			//get direction
			if(i==0){ //first row reached
				left = st.scoringTable[i][j-1];
				max = left;
			}
			else if(j==0){ //first column reached
				top = st.scoringTable[i-1][j];
				max= top;
			}
			else{
				left = st.scoringTable[i][j-1];
				top = st.scoringTable[i-1][j];
				diagonal = st.scoringTable[i-1][j-1];
				
				max = Math.max(left, top);
				max = Math.max(max, diagonal);
			}
			
			if(max==diagonal){
				for(XTrace t: a1.getAlignedTraces()){
					int row = a1.getAlignedTraces().indexOf(t);
					XEvent alignedEventA1 = new XEventImpl();
					alignedEventA1.setAttributes(t.get(i-1).getAttributes()); //create
					alignedAlignmentA1Reverse.get(row).add(alignedEventA1);
//					alignedAlignmentA1Reverse.get(row).getAttributes().put("concept:name", t.getAttributes().get("concept:name"));
				}
				for(XTrace t: a2.getAlignedTraces()){
					int row = a2.getAlignedTraces().indexOf(t);
					XEvent alignedEventA2 = new XEventImpl();
					alignedEventA2.setAttributes(t.get(j-1).getAttributes());
					alignedAlignmentA2Reverse.get(row).add(alignedEventA2);
				}
				
				i -= 1; //move
				j -= 1;
			} 
			else if(max==top){
				for(XTrace t: a1.getAlignedTraces()){
					int row = a1.getAlignedTraces().indexOf(t);
					XEvent alignedEventA1 = new XEventImpl();
					alignedEventA1.setAttributes(t.get(i-1).getAttributes()); //create
					alignedAlignmentA1Reverse.get(row).add(alignedEventA1);
				}
				for(XTrace t: a2.getAlignedTraces()){
					int row = a2.getAlignedTraces().indexOf(t);
					XEvent alignedEventA2 = new XEventImpl();
					XAttributeMap m = new XAttributeMapImpl();
					XAttributeLiteralImpl l = new XAttributeLiteralImpl("concept:name", gapSign);
					m.put("concept:name", l);
					alignedEventA2.setAttributes(m);
					alignedAlignmentA2Reverse.get(row).add(alignedEventA2);;
				}
				i -= 1; //move
			}
			else{
				for(XTrace t: a1.getAlignedTraces()){
					int row = a1.getAlignedTraces().indexOf(t);
					XEvent alignedEventA1 = new XEventImpl();
					XAttributeMap m = new XAttributeMapImpl();
					XAttributeLiteralImpl l = new XAttributeLiteralImpl("concept:name", gapSign);
					m.put("concept:name", l);
					alignedEventA1.setAttributes(m);
					alignedAlignmentA1Reverse.get(row).add(alignedEventA1);
				}
				for(XTrace t: a2.getAlignedTraces()){
					int row = a2.getAlignedTraces().indexOf(t);
					XEvent alignedEventA2 = new XEventImpl();
					alignedEventA2.setAttributes(t.get(j-1).getAttributes()); //create
					alignedAlignmentA2Reverse.get(row).add(alignedEventA2);
				}
				
				j -= 1; //move
			}
			count++;
		}
		
		
		for(XTrace t: alignedAlignmentA1Reverse){
			ArrayList<XEvent> tmpList = new ArrayList<>();
			for(XEvent e: t){
				tmpList.add(e);
			}
			Collections.reverse(tmpList);
			t.clear();
			for(XEvent e: tmpList){
				t.add(e);
			}
		}
		
		for(XTrace t: alignedAlignmentA2Reverse){
			ArrayList<XEvent> tmpList = new ArrayList<>();
			for(XEvent e: t){
				tmpList.add(e);
			}
			Collections.reverse(tmpList);
			t.clear();
			for(XEvent e: tmpList){
				t.add(e);
			}
		}
			
		Alignment alignment = new Alignment("A " +count,a1.getName(), a2.getName(), new ArrayList<XTrace>());
		alignment.getAlignedTraces().addAll(alignedAlignmentA1Reverse);
		alignment.getAlignedTraces().addAll(alignedAlignmentA2Reverse);
		return alignment;
}
	
	/**<p>A function that is used in the construction of the scoring table in the alignAlignments-method.
	 * It influences how the alignment is build, by contributing to the decision whether it would be advantageous for the score to NOT add a gap sign in any of the previous alignments.</p>
	 * 
	 * <p>The event labels at pos1 in a1 are aggregated and compared to the aggregated event labels at pos2 in a2.
	 * </p>
	 * 
	 * @param a1 - An alignment that should be investigated at pos1 in relation to alignment a2 at pos2.
	 * @param gapSign - The string/letter that should be used as a gap sign
	 * 
	 * @return Integer indicating whether not adding a gap-sign is advantageous with this events or not. (high -> rather add, low/negative -> rather not add)
	 */
	private  int subScoreBetweenAlignments(Alignment a1, int pos1, Alignment a2, int pos2){
		
		//different events in column pos1 of a1
		ArrayList<String> distinctEventsA1 = new ArrayList<>();
		
		// frequency of different events in column pos1 of a1
		ArrayList<Integer> n_a1_pos1 = new ArrayList<>();
		
		for(XTrace t: a1.getAlignedTraces()){
			XAttributeLiteralImpl xAtt = (XAttributeLiteralImpl) t.get(pos1-1).getAttributes().get("concept:name");
			String currEvent = xAtt.getValue();
			if(distinctEventsA1.contains(currEvent)){
				int index = distinctEventsA1.indexOf(currEvent);
				n_a1_pos1.set(index, n_a1_pos1.get(index)+1);
			}
			else{
				distinctEventsA1.add(currEvent);
				n_a1_pos1.add(1);
			}
		}
		
		//different events in column pos2 of a2
		ArrayList<String> distinctEventsA2 = new ArrayList<>();
		
		// frequency of different events in column pos2 of a2
		ArrayList<Integer> n_a2_pos2 = new ArrayList<>();
		
		for(XTrace t: a2.getAlignedTraces()){
			XAttributeLiteralImpl xAtt = (XAttributeLiteralImpl) t.get(pos2-1).getAttributes().get("concept:name");
			String currEvent = xAtt.getValue();
			if(distinctEventsA2.contains(currEvent)){
				int index = distinctEventsA2.indexOf(currEvent);
				n_a2_pos2.set(index, n_a2_pos2.get(index)+1);
			}
			else{
				distinctEventsA2.add(currEvent);
				n_a2_pos2.add(1);
			}
		}
		
		//compute score
		int score = 0;
			
			for(int i=0;i<n_a1_pos1.size();i++){
				for(int j=0;j<n_a2_pos2.size();j++){
					
					String labelA1 = distinctEventsA1.get(i);
					String labelA2 = distinctEventsA2.get(j);
					
					int subscore= 0;
					if(!labelA1.equals(gapSign) && !labelA2.equals(gapSign)){
						if(labelA1.equals(labelA2)){ // A A
							subscore = subScoreCase1; //before it was 1
						}
						else{ // A und B
							subscore = subScoreCase2;
						}
					}
					else{ //- - , - A , A -
							subscore = subScoreCase3;
					}
					
					score += n_a1_pos1.get(i)*subscore;
				}
			}
		return score;
	}
	
	/**<p>A function that is used in the construction of the scoring table in the alignAlignments-method.
	 * It influences how the alignment is build, by contributing to the decision whether it would be advantageous for the score to ADD a gap sign in alignment a at position.</p>
	 * 
	 * <p>The event labels at pos1 in a are aggregated and compared to the previous aggregated event labels in a.
	 * </p>
	 * 
	 * @return Integer indicating whether adding a gap-sign is advantageous with this events or not. (high -> rather add, low/negative -> rather not add)
	 */
	private  int indelScoreBetweenAlignments( Alignment a, int pos1){
		
		// overallFrequency of events with nongapSign-predecessors
		int overallFrequency = 0;

		String currEvent2 = "";
		
		for(XTrace t: a.getAlignedTraces()){
			XAttributeLiteralImpl xAtt = (XAttributeLiteralImpl) t.get(pos1-1).getAttributes().get("concept:name");
			String currEvent = xAtt.getValue();
			if(pos1>1){
				XAttributeLiteralImpl xAtt2 = (XAttributeLiteralImpl) t.get(pos1-2).getAttributes().get("concept:name");
				currEvent2 = xAtt2.getValue();
			}
			
			if(!currEvent.equals(gapSign) && 
			 !currEvent2.equals(gapSign)){
				overallFrequency++;
			}
		}
		return overallFrequency*indelScoreCase1;
	}
	
	/**Select a subset from all potential alignments in the input allAlignmentsAtLvl by considering
	 * their evaluation score results. If an alignment pair is taken to the subset, its predecessors are excluded from
	 * the set for building the next alignment pair, because no predecessor should be aligned twice.
	 * 
	 * @param allAlignmentsAtLvl - all POTENTIAL alignments generated in this round
	 * @param evalAtLvl - the evaluation results of the alignments
	 * @param list - list of all EXISTING alignments in total (generated in previous rounds)
	 * @param lvlCounter - the number of rounds
	 * 
	 * @return alignments chosen in this round/ at this hierarchy lvl
	 */
	public  ArrayList<Alignment> findBestAlignsAtCurrentLevel(ArrayList<Alignment> allAlignmentsAtLvl, ArrayList<Integer> evalAtLvl, ArrayList<Alignment> list, int lvlCounter){
		//collect best pairwise alignment and exclude traces which are already in for future comparison
		ArrayList<Alignment> bestAlignments = new ArrayList<>();
		ArrayList<String> exclusion = new ArrayList<>();
		
		int innerCounter = 0;
		
		while(bestAlignments.size()<Math.floor(list.size()/2)){
			int best = Collections.max(evalAtLvl);
			int index = evalAtLvl.indexOf(best);
			
			Alignment bestPwAlign = allAlignmentsAtLvl.get(index);
			evalAtLvl.set(index, -999999); //to avoid running again and again to the same position
			
			if(!exclusion.contains(bestPwAlign.getPredecessor_1()) && !exclusion.contains(bestPwAlign.getPredecessor_2()) ){
				bestPwAlign.setName("A " + lvlCounter +"." + innerCounter);
				innerCounter++;

				bestAlignments.add(bestPwAlign);
				exclusion.add(bestPwAlign.getPredecessor_1());
				exclusion.add(bestPwAlign.getPredecessor_2());
				
			}
		}
		
		if(list.size()>exclusion.size()){
				for(Alignment a: list){
					boolean inExclusion = false;
					String aName = a.getName();
					for(String s: exclusion){
						if(s.equals(aName)){
							inExclusion=true;
						}
					}
					if(!inExclusion){
						bestAlignments.add(a);
						break;
					}
				}
		}
		
		System.out.println();
		System.out.println("BEST ALIGNMENTS AT LEVEL "+ lvl +": ");
		System.out.println();
		
		for(Alignment a : bestAlignments){
			printAlignment(a);
		}
		
	return bestAlignments;
	}

	/**Prints the alignment input.
	 * Used for debugging purposes only.
	 * Invoked at each lvl of the alignment for showing intermediate results.
	 */
	private String printAlignment(Alignment a){

		String alignment="<html> ";

		alignment+=("Parameters: <br>");
		alignment+=("Indelscore: " + this.indelScoreCase1 + " SubScores: " + subScoreCase1 + "," +subScoreCase2 + "," + subScoreCase3 + " EvalScores: " + evalScoreCase1 + "," +evalScoreCase2 + "," + evalScoreCase3 + "," + evalScoreCase4 + "<br>");
		alignment+=("<br>");
		
		
		System.out.println("Parameters: ");
		System.out.println("Indelscore: " + this.indelScoreCase1 + " SubScores: " + subScoreCase1 + "," +subScoreCase2 + "," + subScoreCase3 + " EvalScores: " + evalScoreCase1 + "," +evalScoreCase2 + "," + evalScoreCase3 + "," + evalScoreCase4);
		System.out.println();
		System.out.println("Alignment Candidate "+ a.getName()+ ": " + a.getPredecessor_1() + " with " + a.getPredecessor_2());
		System.out.println();
		
		int i=0;
		for(XTrace t: a.getAlignedTraces()){
			int j=0;
			for(XEvent e: t){
				XAttributeLiteralImpl att = (XAttributeLiteralImpl) e.getAttributes().get("concept:name");
				String label = att.getValue();
				alignment+=label;
				if(label.length()==1){
					System.out.print(label + "   ");
				if(label.equals("-"))
					alignment+="&nbsp;";
				alignment+="&emsp;";
				}
				else if (label.length()==2){
					System.out.print(label + "  ");
					alignment+="&nbsp;&nbsp;";
				}
				else{
					System.out.print(label + " ");
					alignment+="&nbsp;";
				}
				j++;
			}
			System.out.println();
			alignment+=" <br>";
			i++;
		}
		alignment+=" </html>";
		return alignment;
	}

	/**This method rates the alignment of two traces to each other.
	 * In combination with its invocation in the recursiveTA-method,
	 * it evaluates the alignment of multiple traces (in two alignments) by tracewise comparison. 
	 */
	private  int evaluatePartOfAlignment(XTrace t1, XTrace t2){
		
		int result = 0;
		
		for(int i=0;i<t1.size();i++){
			
			//current events at i in T1 & T2
			XAttributeLiteralImpl elemAttT1 = (XAttributeLiteralImpl) t1.get(i).getAttributes().get("concept:name");
			String elemLabelT1 = elemAttT1.getValue();
			XAttributeLiteralImpl elemAttT2 = (XAttributeLiteralImpl)t2.get(i).getAttributes().get("concept:name");
			String elemLabelT2 = elemAttT2.getValue();
			
			//previous events at i-1 in T1 & T2
			String prevElemLabelT1 = null;
			String prevElemLabelT2 = null;
			if(i>0){
				XAttributeLiteralImpl prevElemAttT1 = (XAttributeLiteralImpl) t1.get(i-1).getAttributes().get("concept:name");
				prevElemLabelT1 = prevElemAttT1.getValue();
				XAttributeLiteralImpl prevElemAttT2 = (XAttributeLiteralImpl) t2.get(i-1).getAttributes().get("concept:name");
				prevElemLabelT2 = prevElemAttT2.getValue();		
			}
	
			
			if(!elemLabelT1.equals(gapSign) && !elemLabelT2.equals(gapSign)){
				if(elemLabelT1.equals(elemLabelT2)){
					result += evalScoreCase1; //subScoreCase1;//evalScoreCase1; //A, A
				}
				else{
					result += evalScoreCase2; //subScoreCase2;//evalScoreCase2; // A, B
				}
			}
			else{
//				if  (i>0 && ( (!elemLabelT1.equals(gapSign) && 
//				 !prevElemLabelT1.equals(gapSign) && 
//				  elemLabelT2.equals(gapSign) 
//				 ) ||
//				 (!elemLabelT2.equals(gapSign) && 
//				  !prevElemLabelT2.equals(gapSign) && 
//				  elemLabelT1.equals(gapSign) 
//				 )	)){
						result += evalScoreCase3;  //subScoreCase3;//evalScoreCase3; // A B, -
//				}
//				else{
//					result += evalScoreCase4 ;//TODO understand distinction here
//				}
//	
			}
	
		}
		
		return result;
	}
}

	
