package org.processmining.uncertaintymarker.models;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Map;

import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.impl.XAttributeContinuousImpl;
import org.deckfour.xes.model.impl.XAttributeMapImpl;
import org.processmining.uncertaintymarker.algorithms.Consolidation;
import org.processmining.uncertaintymarker.help.Utilities;

/**A Chunk is used in the {@link Consolidation} to aggregate the confidences of time windows within a log trace which belong together. 
 * These time windows are generated by a classification algorithm in a certain time distance and each consist of a probability distribution representing the confidence.
 * They are grouped and averaged into {@link Alternative}s of a chunk to transfer the automatically generated classification log into a concise log that can be used for process mining techniques.
 * 
 * @author Oliver
 *
 */
public class Chunk{

	String label;
	double avg;
	
	//all probability information from the chunks rows sorted into alternatives.
	ArrayList<Alternative> allAlternatives = new ArrayList<>();

	int rowsInChunk = 0;

	public Chunk() {
	}	
	
	public Chunk(String label, ArrayList<Alternative> list) {
		this.label = label;
		this.allAlternatives = list;
	}
	
	public Chunk(String label){
		this.label = label;
		allAlternatives = new ArrayList<>();
	}

	public double getAvg() {
		return avg;
	}

	public void setAvg(double ownProbAvg) {
		this.avg = ownProbAvg;
	}

	public String getLabel() {
		return label;
	}

	public void setLabel(String label) {
		this.label = label;
	}

	public ArrayList<Alternative> getAllAlternatives() {
		return allAlternatives;
	}

	
	/**Adds all probability information from a row into a chunk.
	 * Updates the number of containted rows and start the computation of the chunks average.
	 */
	public void addItem(XAttributeMapImpl map){
		
		//for each column
		for (Map.Entry<String, XAttribute> entry : map.entrySet()){
			
			if(! (entry.getValue() instanceof XAttributeContinuousImpl))
				continue;

			String alternative = entry.getKey();
			Alternative dummy = new Alternative(alternative, null);
			
			XAttributeContinuousImpl att = (XAttributeContinuousImpl) entry.getValue();
			double value = att.getValue();
			
			//if already in alternatives add prob to list, else new alternative
			if(allAlternatives.contains(dummy)){
				int i = allAlternatives.indexOf(dummy);
				allAlternatives.get(i).getList().add(value);
			}
			else{
				ArrayList<Double> altProbs = new ArrayList<>();
				altProbs.add(value);
				Alternative a = new Alternative(alternative, altProbs);
				allAlternatives.add(a);
			}
		}
		setRowsInChunk(this.getAllAlternatives().get(0).getList().size());
		computeAvg();
		
	}
	
	/** Computes average probability of most probable event Chunk. Invoked after Chunk is full.  */
	public void computeAvg(){
		
		for(Alternative a: allAlternatives){
			a.computeAvg(getRowsInChunk());
		}
		
		Collections.sort(allAlternatives,Collections.reverseOrder());
		
		Alternative max = Collections.max(allAlternatives);
		

		//get secondbest result
		if(!moreThenOneMax()){
			avg = max.getAvg();
			label = max.getLabel();
		}
	}
	
	/**Returns true if more than one max in current row,
	 * else returns false.
	 */
	public boolean moreThenOneMax(){
		
		int count = 0;
		double maxAvg = Collections.max(allAlternatives).getAvg();
		
		for(Alternative a : this.allAlternatives){
			if(a.getAvg()==maxAvg)
				count++;
		}
		
		return count>1;
	}

	public int getRowsInChunk() {
		return rowsInChunk;
	}

	public void setRowsInChunk(int rowsInChunk) {
		this.rowsInChunk = rowsInChunk;
	}
	
	/**Copies alternatives from one chunk  and adds it to another.
	 */
	public void addAlternativesFrom(Chunk c){
		for(Alternative alt : c.allAlternatives){
			for(Alternative alt2  : this.allAlternatives){
				if(alt2.getLabel().equals(alt.getLabel()))
					alt2.getList().addAll(alt.getList());
			}
		}
		
		this.setRowsInChunk(this.allAlternatives.get(0).getList().size());
		this.computeAvg();
	}
	
	@Override
	public String toString(){
		String output = "\n" + this.getLabel() + ", " + Utilities.round(this.getAvg(), 2) + ",rows  " +this.getRowsInChunk();
		return output;
	}
	
}
