package org.processmining.uncertaintymarker.models;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Map;

import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.impl.XAttributeContinuousImpl;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.deckfour.xes.model.impl.XAttributeMapImpl;
import org.processmining.uncertaintymarker.algorithms.AlignedStepDiscovery;
import org.processmining.uncertaintymarker.algorithms.NetAttachment;

/**This class is used in {@link AlignedStepDiscovery} as well as in {@link NetAttachment} 
 * to aggregate confidence and frequency of events over all traces. It is quite similar to {@link Chunk} but
 * has a different addItem-method, a column attribute and is used in another context.
 * 
 * @author Oliver
 *
 */

public class TransitionData{

	String label="";
	double avg=0.0;
	
	ArrayList<Alternative> allAlternatives = new ArrayList<>();

	int frequency = 0;
	
	int column;
	
	/**Adds a event from a column of a certain trace in the aligned log to this object.
	 * They are aggregated in this class for deciding which event is frequently enough in a column
	 * to be included in the net. When it is invoked from Net Attachement, it is collected for later manipulation of the net.
	 */
	public void addItem(XAttributeMapImpl map, int column){
		
		this.column = column;
		
		//get label
		for (Map.Entry<String, XAttribute> entry : map.entrySet()){
			if(entry.getKey().equals("concept:name")){
				XAttributeLiteralImpl xatt = (XAttributeLiteralImpl) entry.getValue();
				label = xatt.getValue();
				break;
			}
		}
		
		//special case of SKIP
		if(label.equals("-")){
			if(allAlternatives.size()<1){
				ArrayList<Double> altProbs = new ArrayList<>();
				altProbs.add(1.0);
				allAlternatives.add(new Alternative(label, altProbs));
			}
			else{
				this.allAlternatives.get(0).getList().add(1.0);
			}
			setFrequency(this.getAllAlternatives().get(0).getList().size());
			computeAvg();
			this.label= allAlternatives.get(0).getLabel();
			return;
		}

		

		//for each column entry in this row
		for (Map.Entry<String, XAttribute> entry : map.entrySet()){


			if(!(entry.getValue() instanceof XAttributeContinuousImpl))
				continue;
			
			String key = entry.getKey();
			XAttributeContinuousImpl tmp = (XAttributeContinuousImpl) entry.getValue();
			double value = tmp.getValue();
			
			Alternative dummy = new Alternative(key, null);

			//if already in alternatives add prob to list, else new alternative
			if(allAlternatives.contains(dummy)){
				int i = allAlternatives.indexOf(dummy);
				allAlternatives.get(i).getList().add(value);
			}
			else{
				ArrayList<Double> altProbs = new ArrayList<>();
				altProbs.add(value);
				Alternative a = new Alternative(key, altProbs);
				allAlternatives.add(a);
			}
			
		}
			
		setFrequency(this.getAllAlternatives().get(0).getList().size());
		computeAvg();
		
		this.label= allAlternatives.get(0).getLabel();
		
	}
	
	/** Computes average probability of most probable event alternative. Invoked after items belonging to this object are added*/
	public void computeAvg(){
		
		for(Alternative a: allAlternatives){
			a.computeAvg(getFrequency());
		}
		
		Collections.sort(allAlternatives,Collections.reverseOrder());
		
		Alternative max = Collections.max(allAlternatives);
		
		avg = max.getAvg();
	
}

	public String getLabel() {
		return label;
	}

	public void setLabel(String label) {
		this.label = label;
	}

	public double getAvg() {
		return avg;
	}

	public void setAvg(double avg) {
		this.avg = avg;
	}

	public ArrayList<Alternative> getAllAlternatives() {
		return allAlternatives;
	}

	public void setAllAlternatives(ArrayList<Alternative> allAlternatives) {
		this.allAlternatives = allAlternatives;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((label == null) ? 0 : label.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Alternative other = (Alternative) obj;
		if (label == null) {
			if (other.label != null)
				return false;
		} else if (!label.equals(other.label))
			return false;
		return true;
	}

	public int getFrequency() {
		return frequency;
	}
	
	public void setFrequency(int freq){
		frequency = freq;
	}
}